# $Id: amazon.rb,v 1.34 2004/03/01 00:51:24 ianmacd Exp $
# 
#:include: ../README.rdoc

module Amazon

  NAME	  = 'Ruby/Amazon'
  VERSION = '0.7.0'

  SITE	  = { 'us' => 'www.amazon.com',
	      'uk' => 'www.amazon.co.uk',
	      'de' => 'www.amazon.de',
	      'jp' => 'www.amazon.co.jp'
	    }

  DEFAULT_ID = 'webservices-20'

  
  # Prints debugging messages and works like printf, except that it prints
  # only when Ruby is run with the -d switch.
  #
  def Amazon.dprintf(format, *args)
    $stderr.printf(format, *args) if $DEBUG
  end


  class Product

    attr_reader :url  # :nodoc:

    #
    # This class holds product reviews.
    #
    class Review

      attr_reader :rating, :summary, :comment

      def initialize(rating, summary, comment)	# :nodoc:
	@rating = rating
	@summary = summary
	@comment = comment
      end

    end


    #
    # This class holds third party information related to offerings.
    #
    class ThirdPartyInfo

      attr_reader :offering_type, :seller_id, :seller_nickname, :exchange_id,
		  :offering_price, :condition, :seller_nickname,
		  :condition_type, :exchange_availability, :seller_country,
		  :seller_state, :ship_comments, :seller_rating

      def initialize  # :nodoc:
      end
    end


    #
    # This exception is raised when an attempt is made to read a product
    # attribute that doesn't exist.
    #
    class AttributeError < StandardError; end


    # 
    # This alias makes the ability to determine a product's properties a
    # little more intuitive.
    #
    alias_method :properties, :instance_variables

    def initialize(url)	# :nodoc:
      @url = url
    end

    #
    # Displays a product in a human-readable format. Call without a parameter,
    # as the parameter is only for internal use.
    # 
    def to_s(first_detail='productname')
      string = ""

      # get all attributes, minus the leading '@'
      vars = self.instance_variables.sort.map {|v| v[1..-1]}

      # find the longest attribute name
      longest = vars.sort { |a,b| b.length <=> a.length }[0].length

      # put the product name at the front of the list, if we have one
      vars.unshift(first_detail) if vars.delete(first_detail)

      # display the product's details
      vars.each do |iv|
	value = self.instance_variable_get("@#{iv}").inspect
	string << "%-#{longest}s = %s\n" % [iv, value]
      end

      string

    end

    #
    # Converts an Amazon::Product to a Hash.
    #
    def to_h
      hash = {}
      self.properties.each do |property|
	key = property[1..-1]
	val = self.instance_variable_get(property)
	hash[key] = val
      end
      hash
    end

    # 
    # Fake the appearance of a product as a hash. _key_ should be any
    # attribute of the product, as returned by Amazon::Product#properties.
    #
    def [](key)
      self.instance_variable_get('@' + key.to_s)
    end

    #
    # Since we can't know in advance which attributes a product will
    # have, we construct a reader method dynamically for each attribute
    # the first time someone tries to read it. So, this method will only
    # be called a maximum of once per attribute for the whole class.
    #
    def method_missing(method)
      iv = "@" + method.id2name

      if instance_variables.include? iv
	Product.module_eval { attr_reader method }
	instance_variable_get iv
      else
        raise AttributeError, "product has no #{iv}"
      end
    end

    private :method_missing

  end


  #
  # This class holds Amazon Seller profile data.
  #
  class Seller < Product
    attr_reader :seller_nickname, :overall_feedback_rating, :nr_feedback,
		:store_id, :store_name, :feedback

    def initialize(nickname, rating, quantity, store_id,
		   store_name, feedback) # :nodoc:
      @seller_nickname = nickname
      @overall_feedback_rating = rating
      @nr_feedback = quantity
      @store_id = store_id
      @store_name = store_name
      @feedback = feedback
    end

    #
    # Convert a seller object to human-readable format.
    #
    def to_s
      super 'store_name'
    end

  end


  #
  # This class holds Amazon Seller feedback.
  #
  class Feedback
    attr_reader :rating, :comments, :date, :rater

    def initialize(rating, comments, date, rater) # :nodoc:
      @rating = rating
      @comments = comments
      @date = date
      @rater = rater
    end

  end


  module Exchange
 
    # Amazon::Exchange::Product
    class Product < Amazon::Product
      def initialize  # :nodoc:
      end
    end

  end


  # This class holds Amazon::ProductLine objects returned by the
  # Amazon::Search::Blended module.
  #
  class ProductLine < Amazon::Product

    attr_reader   :mode, :relevance_rank
    attr_accessor :products

    def initialize(mode, relevance) # :nodoc:
      @mode = mode
      @relevance_rank = relevance
    end

  end

end
