# $Id: search.rb,v 1.97 2004/03/01 03:00:07 ianmacd Exp $
#

require 'amazon'
require 'net/http'
require 'rexml/document'

module Amazon

  # Load this module with:
  #
  #  require 'amazon/search'
  #
  module Search

    HEAVY		= true
    LITE		= false
    LIGHT		= false
    ALL_EDITIONS	= true
    SINGLE_EDITION	= false

    # _RATE_LIMIT_REQUESTS_ must be *true* for compliance with Amazon Web
    # Services regulations, which stipulate no more than one search per
    # second.

    RATE_LIMIT_REQUESTS	= true

    MAX_LITE_ASINS	= 30
    MAX_HEAVY_ASINS	= 10
    MAX_LITE_UPCS	= 30
    MAX_HEAVY_UPCS	= 10
    

    # Returns an Array of valid product search modes, such as:
    #
    # *apparel*, *baby*, *books*, *classical*, *dvd*, *electronics*, *garden*,
    # *kitchen*, *magazines*, *music*, <b>pc-hardware</b> *photo*, *software*,
    # *tools*, *toys*, *universal*, *vhs*, *video*, *videogames*,
    # <b>wireless-phones</b>
    #
    def Search.modes
      %w[apparel baby books classical dvd electronics garden kitchen
	 magazines music pc-hardware photo software tools toys universal vhs
	 video videogames wireless-phones] end


    # Returns an Array of valid offer types, such as:
    #
    # *All*, *ThirdPartyNew*, *Used*, *Collectible*, *Refurbished*
    #
    def Search.offer_types
      %w[All ThirdPartyNew Used Collectible Refurbished]
    end


    # Returns an Array of valid sort types for _mode_, or +nil+ if _mode_
    # is invalid.
    #
    def Search.sort_types(mode)
      sort_types =
	{ 'books'       => %w[+pmrank +salesrank +reviewrank +pricerank
			      +inverse-pricerank +daterank +titlerank
			      -titlerank],

	  'software'    => %w[+pmrank +salesrank +titlerank +price -price],

	  'garden'      => %w[+psrank +salesrank +titlerank -titlerank
			      +manufactrank -manufactrank +price -price],

	  'tools'       => %w[+psrank +salesrank +titlerank -titlerank
			      +manufactrank -manufactrank +price -price],

	  'photo'       => %w[+pmrank +salesrank +titlerank -titlerank],

	  'pc-hardware' => %w[+psrank +salesrank +titlerank -titlerank],

	  'videogames'  => %w[+pmrank +salesrank +titlerank +price -price],

	  'music'       => %w[+psrank +salesrank +artistrank +orig-rel-date
			      +titlerank],

	  'office'      => %w[+pmrank +salesrank +titlerank -titlerank
			      +price -price +reviewrank],

	  'video'       => %w[+psrank +salesrank +titlerank],
				     
	  'electronics' => %w[+pmrank +salesrank +titlerank +reviewrank],
				     
	  'dvd'		=> %w[+salesrank +titlerank],
				     
	  'kitchen'     => %w[+pmrank +salesrank +titlerank -titlerank
			      +manufactrank -manufactrank +price -price],
				     
	  'toys'	=> %w[+pmrank +salesrank +pricerank
			      +inverse-pricerank +titlerank]
	}

      return nil unless sort_types.keys.include? mode

      sort_types[mode]
    end


    # Exception class for failed search requests.
    #
    class SearchError < StandardError; end


    # This is the class around which most others in this library revolve. It
    # contains the most common search methods and exception classes.
    #
    class Request
      attr_reader :token, :id, :locale

      ALL_PAGES	 = 0		# :nodoc:
      AWS_PREFIX = '/onca/xml3' # :nodoc:

      #--
      WEIGHT	 = { HEAVY => 'heavy',
		     LITE  => 'lite'
		   }
      #++

      # Exception class for bad search terms.
      # 
      class TermError < StandardError; end

      # Exception class for bad search modes.
      # 
      class ModeError < StandardError; end

      # Exception class for bad search types.
      #
      class TypeError < StandardError; end

      # Exception class for bad offer types.
      #
      class OfferError < StandardError; end

      # Exception class for bad locales.
      #
      class LocaleError < StandardError; end

      # Exception class for bad sort types.
      #
      class SortError < StandardError; end

      # Exception class for HTTP errors (anything other than <b>200</b>)
      #
      class HTTPError < StandardError; end


      def initialize(dev_token, assoc_id=DEFAULT_ID, l='us',
		     user_agent = "%s %s" % [NAME, VERSION])

	locale = { 'us' => 'xml.amazon.com',
		   'uk' => 'xml-eu.amazon.com',
		   'de' => 'xml-eu.amazon.com',
		   'jp' => 'xml.amazon.co.jp',
		 }
	l.downcase!

	unless locale.keys.include? l
	  raise LocaleError, 'invalid locale'
	end

	@token  = dev_token
	@id     = assoc_id || DEFAULT_ID
	@conn   = Net::HTTP.start(locale[l])
	@locale	= l
	@user_agent = user_agent
      end


      # Encode a string, such that it is suitable for HTTP transmission.
      #
      def url_encode(string)
	# shamelessly plagiarised from Wakou Aoyama's cgi.rb
	string.gsub(/([^ a-zA-Z0-9_.-]+)/n) do
          '%' + $1.unpack('H2' * $1.size).join('%').upcase
        end.tr(' ', '+')
      end
      private :url_encode


      # Convert a US mode string into a localised mode string.
      # 
      # English-language books have their own special mode on non-English
      # speaking sites, so we use the pseudo-mode 'english-books' for the US
      #
      def localise_mode(m)

	if @locale == 'us' && m == 'english-books'
	  raise ModeError, "Invalid mode '#{m}' in locale '#{@locale}'"
	end

	return m if @locale == 'us'

	mode = { 'books'	  => { 'uk' => 'books-uk',
				       'de' => 'books-de',
				       'jp' => 'books-jp'
				     },
		 'music'	  => { 'uk' => 'music',
				       'de' => 'pop-music-de',
				       'jp' => 'music-jp'
				     },
		 'classical'	  => { 'uk' => 'classical',
				       'de' => 'classical-de',
				       'jp' => 'classical-jp'
				     },
		 'dvd'		  => { 'uk' => 'dvd-uk',
				       'de' => 'dvd-de',
				       'jp' => 'dvd-jp'
				     },
		 'vhs'		  => { 'uk' => 'vhs-uk',
				       'de' => 'vhs-de',
				       'jp' => 'vhs-jp'
				     },
		 'video'	  => { 'uk' => 'video',
				       'de' => 'video',
				       'jp' => 'video-jp'
				     },
		 'electronics'	  => { 'uk' => 'electronics',
				       'de' => 'ce-de',
				       'jp' => 'electronics-jp'
				     },
		 'kitchen'	  => { 'uk' => 'kitchen-uk',
				       'de' => 'kitchen-de',
				       'jp' => nil
				     },
		 'software'	  => { 'uk' => 'software-uk',
				       'de' => 'software-de',
				       'jp' => 'software-jp'
				     },
		 'videogames'	  => { 'uk' => 'video-games-uk',
				       'de' => 'video-games-de',
				       'jp' => 'videogames-jp'
				     },
		 'magazines'	  => { 'uk' => nil,
				       'de' => 'magazines-de',
				       'jp' => nil
				     },
		 'toys'		  => { 'uk' => 'toys-uk',
				       'de' => nil,
				       'jp' => nil
				     },
		 'photo'	  => { 'uk' => nil,
				       'de' => nil,
				       'jp' => nil
				     },
		 'baby'		  => { 'uk' => nil,
				       'de' => nil,
				       'jp' => nil
				     },
		 'garden'	  => { 'uk' => nil,
				       'de' => nil,
				       'jp' => nil
				     },
		 'pc-hardware'	  => { 'uk' => nil,
				       'de' => nil,
				       'jp' => nil
				     },
		 'tools'	  => { 'uk' => nil,
				       'de' => nil,
				       'jp' => nil
				     },
		 'english-books'  => { 'uk' => nil,
				       'de' => 'books-de-intl-us',
				       'jp' => 'books-us'
				     }
	       }

	if mode[m][@locale].nil?
	  raise ModeError, "Invalid mode '#{m}' in locale '#{@locale}'"
	end

	mode[m][@locale]
      end
      private :localise_mode


      # Deal with a request for offers.
      #
      def handle_offers(url, offer=nil)
	unless offer.nil?
	  if @locale == 'de'
	    raise LocaleError, "search type invalid in '#{@locale}' locale"
	  end

	  unless Search.offer_types.include? offer
	    raise OfferError, "'offerings' must be one of %s" %
			      Search.offer_types.join(', ')
	  end

	  @type = WEIGHT[HEAVY]
	  url << "&offer=" << offer
	end

	url
      end
      private :handle_offers


      # Deal with a request for a particular sort type.
      #
      def handle_sorts(url, sort_type, mode)

	unless sort_type.nil?
	  unless Search.sort_types(mode).include? sort_type
	    raise SortError,
	      "invalid sort type '#{sort_type}' for mode #{mode}"
	  end

	  url << "&sort=" << url_encode(sort_type)
	end

	url
      end
      private :handle_sorts


      # Search for a product by keyword(s) and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def keyword_search(keyword, mode='books', weight=HEAVY, page=1,
			 sort_type=nil, offerings=nil, price=nil,
			 editions=SINGLE_EDITION, &block)

	url = AWS_PREFIX + "?t=%s&KeywordSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&price=" << price unless price.nil?
	url << "&variations=yes"  if	 editions == ALL_EDITIONS
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	keyword = url_encode(keyword)

	unless Search.modes.include? mode
	  raise ModeError, "mode must be one of %s" % Search.modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, keyword, mode, @type, @token, page], &block)
      end


      # Search for a product by browse node. The default of '1000' is for
      # best-selling books. Returns an Amazon::Search::Response. If a block is
      # given, that Response's @products will be passed to the block.
      #
      def node_search(browse_node='1000', mode='books', weight=HEAVY, page=1,
		      sort_type=nil, offerings=nil, keyword=nil, price=nil,
		      &block)

	url = AWS_PREFIX + "?t=%s&BrowseNodeSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	if browse_node.is_a? Array
	  raise TypeError, "string or integer required"
	elsif browse_node =~ / /
	  raise TermError, "single item expected"
	end

	unless Search.modes.include? mode
	  raise ModeError, "mode must be one of %s" % Search.modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, browse_node, mode, @type, @token, page], &block)
      end


      # Search for a product by ASIN(s) and returns an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block. The _offer_page_ parameter is
      # ignored unless _offerings_ is not +nil+.
      #
      def asin_search(asin, weight=HEAVY, offer_page=nil, offerings=nil,
		      &block)

	url = AWS_PREFIX + "?t=%s&AsinSearch=%s&f=xml&type=%s&dev-t=%s"
 	@type = WEIGHT[weight]

	unless offerings.nil?
	  url = handle_offers(url, offerings)
	  url << "&offerpage=%s" % (offer_page || 1)
	end

	asin.gsub!(/ /, ',') if asin.is_a? String
	asin = asin.join(',') if asin.is_a? Array

	if asin.count(',') >= (weight ? MAX_HEAVY_ASINS : MAX_LITE_ASINS)
	  raise TermError, "too many ASINs"
	end

	search(url % [@id, asin, @type, @token], &block)
      end


      # Search for a product by UPC code(s) and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def upc_search(upc, mode='music', weight=HEAVY, &block)

	unless @locale == 'us'
	  raise LocaleError, "search type invalid in '#{@locale}' locale"
	end

	url = AWS_PREFIX + "?t=%s&UpcSearch=%s&mode=%s&f=xml&type=%s&dev-t=%s"
 	@type = WEIGHT[weight]

	upc.gsub!(/ /, ',') if upc.is_a? String
	upc = upc.join(',') if upc.is_a? Array

	if upc.count(',') >= (weight ? MAX_HEAVY_UPCS : MAX_LITE_UPCS)
	  raise TermError, "too many UPCs"
	end

	modes = %w[music classical software dvd vhs video
		   electronics pc-hardware photo]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, upc, mode, @type, @token], &block)
      end

      
      # Search for a book by author and return an Amazon::Search::Response. If
      # a block is given, that Response's @products will be passed to the
      # block.
      #
      def author_search(author, mode='books', weight=HEAVY, page=1,
			sort_type=nil, offerings=nil, keyword=nil,
			price=nil, editions=SINGLE_EDITION, &block)

	url = AWS_PREFIX + "?t=%s&AuthorSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
	url << "&variations=yes"       if editions == ALL_EDITIONS
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

 	@type = WEIGHT[weight]
	author = url_encode(author)

	raise ModeError, 'mode must be books' unless mode == 'books'

	mode = localise_mode(mode)
	search(url % [@id, author, mode, @type, @token, page], &block)
      end

      
      # Search for a book, using a power search, and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def power_search(query, mode='books', weight=HEAVY, page=1,
		       sort_type=nil, offerings=nil, editions=SINGLE_EDITION,
		       &block)

	url = AWS_PREFIX + "?t=%s&PowerSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%"
	url << "&variations=yes" if editions == ALL_EDITIONS
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

 	@type = WEIGHT[weight]
	query = url_encode(query)

	raise ModeError, 'mode must be books' unless mode == 'books'

	mode = localise_mode(mode)
	search(url % [@id, query, mode, @type, @token, page], &block)
      end

      
      # Search for a product by artist and return an Amazon::Search::Response.
      # If a block is given, that Response's @products will be passed to the
      # block.
      #
      def artist_search(artist, mode='music', weight=HEAVY, page=1,
			sort_type=nil, offerings=nil, keyword=nil,
			price=nil, &block)

	url = AWS_PREFIX + "?t=%s&ArtistSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	artist = url_encode(artist)

	modes = %w[music classical]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, artist, mode, @type, @token, page], &block)
      end

      
      # Search for a product by actor and return an Amazon::Search::Response.
      # If a block is given, that Response's @products will be passed to the
      # block.
      #
      def actor_search(actor, mode='dvd', weight=HEAVY, page=1,
		       sort_type=nil, offerings=nil, keyword=nil,
		       price=nil, &block)

	url = AWS_PREFIX + "?t=%s&ActorSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	actor = url_encode(actor)

	modes = %w[dvd vhs video]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, actor, mode, @type, @token, page], &block)
      end

      
      # Search for a product by director and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def director_search(director, mode='dvd', weight=HEAVY, page=1,
			  sort_type=nil, offerings=nil, keyword=nil,
			  price=nil, &block)

	url = AWS_PREFIX + "?t=%s&DirectorSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	director = url_encode(director)

	modes = %w[dvd vhs video]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, director, mode, @type, @token, page], &block)
      end

      
      # Search for a product by manufacturer and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def manufacturer_search(director, mode='electronics', weight=HEAVY,
			      page=1, sort_type=nil, offerings=nil,
			      keyword=nil, price=nil, &block)

	url = AWS_PREFIX + "?t=%s&ManufacturerSearch=%s&mode=%s" +
	      "&f=xml&type=%s&dev-t=%s&page=%s"
	url << "&keywords=" << keyword unless keyword.nil?
	url << "&price="    << price   unless price.nil?
 	@type = WEIGHT[weight]
	url = handle_offers(url, offerings)
	url = handle_sorts(url, sort_type, mode)

	director = url_encode(director)

	modes = %w[electronics kitchen videogames software
		   photo pc-hardware]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	mode = localise_mode(mode)
	search(url % [@id, director, mode, @type, @token, page], &block)
      end

      
      # Return an Amazon::Search::Response of the products on a Listmania list.
      # If a block is given, that Response's @products will be passed to the
      # block.
      #
      def listmania_search(list_id, weight=HEAVY, &block)

	url = AWS_PREFIX + "?t=%s&ListManiaSearch=%s&f=xml&type=%s&dev-t=%s"
 	@type = WEIGHT[weight]

	unless list_id.length.between?(12, 13)
	  raise TermError, "list ID length must be 12 or 13 characters"
	end

	search(url % [@id, list_id, @type, @token], &block)
      end

      
      # Search for a product's similar products and return an
      # Amazon::Search::Response. If a block is given, that Response's
      # @products will be passed to the block.
      #
      def similarity_search(asin, weight=HEAVY, page=1, &block)

	url = AWS_PREFIX + "?t=%s&SimilaritySearch=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
 	@type = WEIGHT[weight]

	asin.gsub!(/ /, ',') if asin.is_a? String
	asin = asin.join(',') if asin.is_a? Array

	if asin.count(',') >= 5
	  raise TermError, "too many ASINs (max. 5 for this search)"
	end

	search(url % [@id, asin, @type, @token, page], &block)
      end


      # Return an Amazon::Search::Response of the products on a wishlist.
      # If a block is given, that Response's @products will be passed to the
      # block.
      #
      def wishlist_search(list_id, weight=HEAVY, &block)

	url = AWS_PREFIX + "?t=%s&WishlistSearch=%s&f=xml" +
	      "&type=%s&dev-t=%s"
 	@type = WEIGHT[weight]

	unless list_id.length.between?(12, 13)
	  raise TermError, "list ID length must be 12 or 13 characters"
	end

	search(url % [@id, list_id, @type, @token], &block)
      end


      # Perform a text stream search and return an Amazon::Search::Response.
      # If a block is given, that Response's @products will be passed to the
      # block.
      #
      def text_stream_search(text_stream, mode='books', weight=HEAVY,
			     sort_type=nil, page=1, &block)
	    
        # this search type not available for international sites
        unless @locale == 'us'
	  raise LocaleError, "search type invalid in '#{@locale}' locale"
        end
  
        url = AWS_PREFIX + "?t=%s&TextStreamSearch=%s&mode=%s&f=xml" +
	      "&type=%s&dev-t=%s&page=%s"
	@type = WEIGHT[weight]
	url = handle_sorts(url, sort_type, mode)

	modes = %w[electronics books videogames apparel toys photo
		   music dvd wireless-phones]
	unless modes.include? mode
	  raise ModeError, "mode must be one of %s" % modes.join(', ')
	end

	# strip a few useless words from the text stream
	%w[and or not the a an but to for of on at].each do |particle|
	  text_stream.gsub!(/\b#{particle}\b/, '')
	end

	text_stream = url_encode(text_stream)
	search(url % [@id, text_stream, mode, @type, @token, page], &block)
      end


      # Handle non-existent and unimplemented search types.
      #
      def method_missing(*params)
	raise TypeError,
	      "non-existent/unimplemented search type: #{params[0].id2name}"
      end
      private :method_missing


      # Get page, but throw exception if there's an HTTP error
      #
      def get_page(url)	  # :nodoc:
	response = @conn.get(url, { 'user-agent', @user_agent })

	if response.code != '200'
	  raise HTTPError, "HTTP response code #{response.code}"
	end

	response.body
      end


      # Perform the actual search.
      #
      def search(url, &block) # :nodoc:
	url << "&locale=" << @locale

	if url.sub!(/page=#{ALL_PAGES}/, 'page=1')
	  # we need to retrieve all pages
	  all_pages = true
	else
	  all_pages = false
	end

	Amazon::dprintf("Fetching http://%s%s...\n", @conn.address, url)

	# get the page
	body = get_page(url)

	body = case self
	       when Amazon::Search::Exchange::Marketplace::Request
		 Exchange::Marketplace::Response.new(body)

	       when Amazon::Search::Exchange::ThirdParty::Request
		 Exchange::ThirdParty::Response.new(body)

	       when Amazon::Search::Exchange::Request
		 Exchange::Response.new(body)

	       when Amazon::Search::Blended::Request
		 Blended::Response.new(body)

	       when Amazon::Search::Seller::Request
		 Seller::Response.new(body)

	       when Amazon::ShoppingCart
		 body

	       else   # must be Amazon::Search::Request
		 Response.new(body)
	       end

	# FIXME: this is a gross little hack to return wishlists in
	# descending order.
	#
	if caller[0] =~ /`wishlist_search'$/
	  body.products = body.products.reverse
	end

	if all_pages
	  responses = [body]
	  threads = []

	  # Get second and subsequent pages in parallel
	  # FIXME: why is this not faster?
	  #
	  2.upto(body.products.total_pages) do |page_nr|

	    # be nice to Amazon
	    sleep 1 if RATE_LIMIT_REQUESTS &&
		      ! ENV.has_key?('AMAZON_NO_RATE_LIMIT')

	    threads << Thread.new(url) do |paged_url|

	      req = Request.new(@token, @id, @locale, @user_agent)
	      
	      # increment page number
	      paged_url.sub!(/page=\d+/) { |s| "page=#{page_nr}" }

	      # go on and get next body, appending to our list
	      response = Response.new(req.get_page(paged_url))
	      responses << response
	    end
	  end
	  threads.each { |t| t.join }

	  if responses.size != body.products.total_pages
	    raise SearchError, "Failed to get all pages"
	  end

	  # Define a singleton method for this Array, so that we can retrieve
	  # all products of each Response in a single method call. I.e.
	  # there's no need to do something like this:
	  #
	  # responses.each { |r| r.products.each { |p| puts p } }
	  #
	  # Instead, we can do this:
	  #
	  # responses.products.each { |p| puts p }
	  #
	  def responses.products
	    products = []
	    each { |page| products << page.products }
	    products.flatten!
	  end

	  # return an Array of Responses, sorted on page number
	  responses.sort! { |a,b| a.args['page'].to_i <=> b.args['page'].to_i }

	  responses.products.each(&block) if block_given?
	  responses

	else
	  # return a single Response
	  body.products.each(&block) if block_given?
	  body
	end
      end
      private :search

    end
    

    class Response < String

      attr_reader :stream	      # :nodoc:
      attr_reader :args

      #--
      # FIXIT: This can become a reader method if wishlists ever revert to
      # being returned in descending order
      #++
      attr_accessor :products

      def initialize(stream)
	@args = {}
	@error = nil
	@stream = nil

	if stream.is_a? File
	  # we were passed an open file handle -- slurp it as a string
	  @stream = stream
	  super stream.readlines(nil)[0]
	  @stream.close
	elsif stream.is_a? REXML::Element
	  @stream = stream
	else  # String
	  super stream
	end

	parse
	
      end


      # Parse the request/response arguments.
      #
      def get_args(node, detail_node=node)  # :nodoc:
        node.elements.each('Request/Args/Arg') do |arg|
          key = arg.attributes['name']
          val = arg.attributes['value']
          @args[key] = val
        end

	Amazon::dprintf("Response args = %s\n", @args.inspect)

	# Check for the presence of actual results.
	#
	unless detail_node.has_elements?
	  @error = "empty result set"
	  raise SearchError, @error
	end

	begin
	  if node.elements['ErrorMsg'].nil?
	    @error = detail_node.elements['ErrorMessage'].text
	  else
	    @error = detail_node.elements['ErrorMsg'].text
	  end

	  raise SearchError, @error

	rescue NoMethodError
	  # AWS has not reported an error on their side.
	end

	if @args.empty? && ! @stream.is_a?(File)
	  raise SearchError, "response contained no arguments"
	end

      end
      
      
      # Parse an XML Amazon::Search::Request and return an
      # Amazon::Search::Response.
      #
      def parse
	@products = []

	# create a singleton #inspect for looking at this Array, including
	# its instance variables for TotalResults and TotalPages.
	#
	class << @products
	  attr_accessor :total_results, :total_pages
	  alias_method	:old_inspect, :inspect

	  def inspect

	    str = ""

	    unless @total_pages.nil?
	      str << "@total_pages=#{@total_pages}, " +
	      str << "@total_results=#{@total_results},\n"
	    end
	    str << old_inspect
	  end
	end

	if @stream.nil? || @stream.is_a?(File)
	  doc = REXML::Document.new(self).elements['ProductInfo']

	  # populate @args with header data
	  get_args(doc) if @args.empty?
	else
	  doc = @stream
	end

	begin
	  @products.total_results = doc.elements['TotalResults'].text.to_i
	  @products.total_results.freeze
	  @products.total_pages = doc.elements['TotalPages'].text.to_i
	  @products.total_pages.freeze
	rescue
	  @products.total_results = nil
	  @products.total_pages = nil
	end

	doc.elements.each('Details') do |detail|
	  product = Product.new(detail.attributes['url'])

	  detail.elements.each do |property|

	    if property.has_elements?

	      case property.name
	      # deal with elements that have more than one sub-level
		
	      when 'BrowseList'
		browsenames = property.elements.map do |e|
				e.elements.map { |e| e.text }
			      end.flatten
		product.instance_variable_set(:@browselist, browsenames)

	      when 'Reviews'
		# can be either AverageCustomerRating or AverageRating
		avg = property.elements[1].text
		tcr = property.elements[2].text

		list =
		  property.elements.map { |e| e.elements.map { |e| e.text } }
		reviews = []
		list.each do |r|
		  reviews << Product::Review.new(*r) unless r.empty?
		end
		product.instance_variable_set(:@avgcustomerrating, avg)

		if tcr.to_i > 0	  # not all review sets have this attribute
		  product.instance_variable_set(:@totalcustomerreviews, tcr)
		end

		product.instance_variable_set(:@reviews, reviews)

	      when 'ThirdPartyProductInfo'  # Offerings returns these

		info = []
		property.elements.map do |e|
		  tpi = Product::ThirdPartyInfo.new
		  e.elements.map do |e|
		    iv, value = e.name, e.text

		    unless value.nil?
		      tpi.instance_variable_set(
			"@#{iv.downcase}".to_sym, value)
		    end

		  end

		  info << tpi
		end

		product.instance_variable_set(:@thirdpartyproductinfo, info)

	      else  # deal with the rest
		members = property.elements.map { |e| e.text }

		# better than Object#instance_eval, but still
		product.instance_variable_set(
		  "@#{property.name.downcase}".to_sym, members)
	      end

	    else	# these elements have no children

	      # more ugliness
	      product.instance_variable_set(
		"@#{property.name.downcase}".to_sym, property.text)
	    end

	  end

	  @products << product

	end

	self

      end
      private :parse

    end

  end
end

require 'amazon/search/blended'
require 'amazon/search/exchange'
require 'amazon/search/exchange/marketplace'
require 'amazon/search/exchange/thirdparty'
require 'amazon/search/seller'
require 'amazon/shoppingcart'
